use rustracing_jaeger::Span;

use crate::{
    db::{DBConnection, UserRole},
    metrics,
    routes::{auth, AppError}
};

pub type UserInfo = (crate::db::User, crate::db::Token);

pub fn authorize_jwt(span: &Span, jwt: &str, db: &mut DBConnection) -> Result<UserInfo, AppError> {
    let inner_span = metrics::span("authorize_jwt", span);
    let decoded = jsonwebtoken::decode::<auth::JWTClaims>(
        jwt,
        &auth::JWT_DECODE_KEY,
        &jsonwebtoken::Validation::new(auth::JWT_ALGORITHM)
    )
    .map_err(|_| AppError::Forbidden("Invalid token"))?;

    db.cleanup_tokens(span);

    let mut user = db.get_user(&inner_span, decoded.claims.sub).ok_or(AppError::Forbidden("Invalid token"))?;
    let token = db.get_token(&inner_span, decoded.claims.jti).ok_or(AppError::Forbidden("Invalid token"))?;

    if user.id != token.owner_id {
        return AppError::Forbidden("Invalid token").err();
    }
    if user.role == UserRole::Disabled {
        return AppError::Forbidden("Account disabled").err();
    }
    if user.gitlab {
        let info = auth::gitlab::get_gitlab_user(&inner_span, user.gitlab_at.clone().unwrap());
        let info = match info {
            Some(v) => Some(v),
            None => {
                let tokens = auth::gitlab::get_gitlab_token(&inner_span, user.gitlab_rt.clone().unwrap(), true);
                if let Some(tokens) = tokens {
                    user.gitlab_at = Some(tokens.access_token.clone());
                    user.gitlab_rt = Some(tokens.refresh_token);
                    db.save_user(&inner_span, &user);
                    auth::gitlab::get_gitlab_user(&inner_span, tokens.access_token)
                } else {
                    None
                }
            }
        };
        if info.is_none() || info.unwrap().username != user.name {
            db.delete_all_tokens(&inner_span, token.owner_id);
            db.delete_all_tokens(&inner_span, user.id);
            return AppError::Forbidden("Invalid gitlab user").err();
        }
    }

    Ok((user, token))
}
